#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

from math import *
from string import *
import re
import random
import operator

# fitter = FuzzyFitter()
# fitter.setEqn('M * x + B')
# or fitter.setFunc(lambda M,B,x: M*x+B, 'M*x+B', ['M', 'B'])    # please make x the final argument
# fitter.setComponentCount(1)
# fitter.useData( [1, 2, 3], [5, 3, 1], decimation=2 ) # using every 2nd point
# fitter.fit() # optional/named args; returns number of iterations
# fitter.ssd() # returns sum-square-difference between curves and data
# also see various methods to get/set parameters positionally or by name
# and baseline: subtracted from each y


class FuzzyFitter:
    def __init__(self):
        self.ncomponent = 1
        self.setEqn('Amp * exp( - x / Tau )')
        self.X = [1]
        self.Y = [1]
        self.baseline = 0.0

    def setFunc(self, func, eqn, argnames):
        self.func = func
        self.eqn = eqn
        self.paramNames = argnames[:]
        self.params = {}
        for i in range(len(argnames)):
            self.params[argnames[i]] = i
        pvalsc = [1.0] * len(self.params.keys())
        self.pvals = []
        for c in range(self.ncomponent):
            self.pvals.append( pvalsc[:] )
        

    def setEqn(self, neweqn):
        args = []
        valid = 0
        newfunc = lambda : 0.0

        def MissingName(msg): # extract the unknown name from a NameError message
            m = re.search("'([^']+)'", msg)
            if not m:
                raise NameError, msg
            return m.group(1)

        while not valid:
            try:
                newfunc = eval( 'lambda ' + join(args+['x'], ',') + ': ' + neweqn )
                fval = float( apply( newfunc, [1.0]*(len(args)+1) ) )
                valid = 1
            except NameError, msg:
                args.append( MissingName(str(msg)) )
        
        self.setFunc(newfunc, neweqn, args)

    def componentCount(self):
        return self.ncomponent

    def setComponentCount(self, n):
        if self.ncomponent > n:
            self.ncomponent = n
            self.pvals = self.pvals[:n]
        while self.ncomponent < n:
            self.ncomponent = self.ncomponent + 1
            self.pvals.append( self.pvals[-1] )
    
    def paramCount(self):
        return len(self.pvals[0])
    def paramName(self, i):
        return self.paramNames[i]

    def param(self, c, i):
        return self.pvals[c][i]
    def paramByName(self, c, name):
        return self.pvals[c][self.params[name]]
    def setParam(self, c, i, x):
        self.pvals[c][i] = x
    def setParamByName(self, c, name, x):
        self.pvals[c][self.params[name]] = x

    def setBaseline(b):
        self.baseline = b

    def useData(self, X, Y, decimation=1):
        if decimation <= 1:
            self.X = X
            self.Y = Y
        else:
            self.X = []
            self.Y = []
            for i in range(0, min(len(X), len(Y)), decimation):
                self.X.append(X[i])
                self.Y.append(Y[i])

    def ssd(self, altPvals=None):
        if altPvals:
            savePvals = self.pvals
            self.pvals = altPvals
            
        Y = self.Y
        X = self.X
        N = min(len(Y), len(X))
        ncomponent = self.componentCount()

        def makeComponentFunc(func, pvalsc):
            return lambda x: apply(func, pvalsc+[x])
        comp_funcs = [makeComponentFunc(self.func, self.pvals[c]) for c in range(ncomponent)]
        baseline = self.baseline

        def sd1Pt(i):
            fitval = sum([f(X[i]) for f in comp_funcs])
            diff = Y[i] - baseline - fitval
            return diff * diff
        result = sum(map(sd1Pt, xrange(N)))

        #result = 0.0
        #for i in xrange(N):
        #    fitval = 0.0
        #    for c in range(ncomponent):
        #        fitval = fitval + comp_funcs[c](X[i])
        #    diff = Y[i] - baseline - fitval
        #    result = result + diff * diff

        if altPvals:
            self.pvals = savePvals

        return result

    def fit(self, mag=0.05, nfuzz=10, maxnfuzz=20, maxiter=100, precision=0.1, onIter=lambda x:x): # returns number of iterations
        def fuzzone(val):
            return val + val * mag * (2.0*random.random()-1.0)
        def fuzzcomp(pvalrow):
            return [fuzzone(x) for x in pvalrow]
        def fuzzout(pvals):
            return [fuzzcomp(x) for x in pvals]

        def fuzzbatch(pvals, count):
            pvalsOfSSD = {}
            for i in range(count):
                fpvals = fuzzout(pvals)
                fssd = self.ssd(fpvals)
                pvalsOfSSD[fssd] = fpvals
            return pvalsOfSSD

        ssd_here = self.ssd(self.pvals)
        last_iter = 0
        for iter in range(maxiter):
            last_iter = iter
            pvalsOfSSD = fuzzbatch(self.pvals, nfuzz)
            bestssd = min(pvalsOfSSD.keys())
            if (bestssd >= ssd_here) and (maxnfuzz > nfuzz):
                pvalsOfSSD = fuzzbatch(self.pvals, maxnfuzz - nfuzz)
                bestssd = min(pvalsOfSSD.keys())
            if bestssd >= (ssd_here - precision):
                break
            self.pvals = pvalsOfSSD[bestssd]
            ssd_here = bestssd
            onIter(ssd_here)

        return last_iter 


def concat(lsts):
    return reduce(lambda x,y:x+y, lsts)

def cartprod(*lsts):
    def extend(prod, remains):
        if 0 == len(remains):
            return [prod]
        subremains = remains[1:]
        return concat([extend(prod+[el], subremains) for el in remains[0]])
    return extend([], lsts)

def hsphereToCartesian(rad, angles):
    root = rad
    cart = [0.0] * (len(angles) + 1)
    for i in xrange(len(angles)):
        cart[i] = root * cos(angles[i])
        root = root * sin(angles[i])
    cart[-1] = root
    return cart
def cleantup(digits):
    def cleantupd(tup):
        return tuple([round(x, digits) for x in tup])
    return cleantupd
hclean = cleantup(7)

def hypersphere(N, ncircle, rad=1.0, rand=True):
    step = 2*pi/ncircle
    steps = [i*step for i in xrange(ncircle)]
    if rand:
        base = [2*pi*random.random() for i in xrange(N-1)]
    else:
        base = [0.0] * (N-1)
    angles = [ [b+s for s in steps] for b in base ]
    for hcoord in cartprod(*angles):
        yield hclean(hsphereToCartesian(rad, hcoord))

def unique(seq):
    seen = {}
    for x in seq:
        if not seen.has_key(x):
            seen[x] = True
            yield x

class DumbGradFitter(FuzzyFitter):
    def fit(self, rad=0.5, decay=0.9, ncircle=100, linmean=1.0, lincount=100, maxiter=100, precision=0.1, onIter=lambda x:x): # returns number of iterations

        npar = sum([len(pval) for pval in self.pvals])
        #hsphere = [x for x in unique(hypersphere(npar, ncircle, 1.0, True))]

        def radsample(rad):
            return hsphereToCartesian(rad, [random.random()*2*pi for i in xrange(npar-1)])
        def radsamples(rad, count):
            for i in xrange(count):
                yield radsample(rad)

        def fuzzone(val):
            return val * rad * (2.0*random.random()-1.0)
        def fuzzcomp(pvalrow):
            return [fuzzone(x) for x in pvalrow]
        def fuzzout(pvals):
            return reduce(lambda x,y:x+y, [fuzzcomp(x) for x in pvals], [])
        def fuzzsamples(pvals, count):
            for i in xrange(count):
                yield fuzzout(pvals)

        def seqsearch(pvals, ssd):
            best = None
            for pval in pvals:
                pssd = self.ssd(pval)
                if pssd < ssd:
                    ssd = pssd
                    best = pval
            return best, ssd
        def circle(pvals, rad):#, hsphere):
            for deltas in radsamples(rad, ncircle):#hsphere:
                next = []
                i = 0
                for component in pvals:
                    out_comp = []
                    for par in component:
                        out_comp.append(par+deltas[i])
                        i += 1
                    next.append(out_comp)
                yield next
        def line(pvals, direction, linmean, lincount):
            for i in xrange(lincount):
                step = random.expovariate(linmean)
                yield [ [p+step*d for p, d in zip(pval, dir)] for pval, dir in zip(pvals, direction) ]
        ssd_here = self.ssd(self.pvals)
        last_iter = 0
        for iter in range(maxiter):
            last_iter = iter
            circle_best, circle_ssd = seqsearch(circle(self.pvals, rad), ssd_here)
            rad *= decay
            if circle_best:
                direction = [ [b-p for b,p in zip(bval,pval)] for bval, pval in zip(circle_best, self.pvals)]
                lin_best, lin_ssd = seqsearch(line(self.pvals, direction, linmean, lincount), ssd_here)
                #if lin_ssd < (ssd_here + precision):
                #    break
                if lin_best:
                    self.pvals = lin_best
                    ssd_here = lin_ssd
            onIter(ssd_here)
        return last_iter 

